# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/10_Download-demo-data.ipynb.

# %% auto 0
__all__ = ['download']

# %% ../notebooks/10_Download-demo-data.ipynb 13
import wget 
import re 
import os 
import sys
import ssl 
import filehash 

def _bar_progress(current, total, width=80):
    '''Create this bar_progress method which is invoked automatically from wget'''

    percent = (100 * current) // total 
    progress_message = f'Downloading {total//1e6} MB: {percent}% '
    # Don't use print() as it will print in new line every time.
    sys.stdout.write("\r" + progress_message)
    sys.stdout.flush()

def download():
    '''Download demo files from cloud storage. '''
    
    # trying to fix SSL errors with wget.download()  below 
    ssl._create_default_https_context = ssl._create_unverified_context
    
    cwd = os.getcwd()

    bucket_url = 'https://f002.backblazeb2.com/file/crono-maxrf-data-demo/'

    fname_list = ['14200215102021-blindTest2AgedDetail.HDF5', 
                  'Ink-08-aged-VIS.png'] 
    
    #sha1 file hashes 
    fhash_list = ['386bff97bed28d0302bf93abd00d591c4f45b44d', 
                 '7d8c7f1c5f105f10c0ec06fdb9440cf3ea9c00aa']

    n_files = len(fname_list)


    print(f'Checking your current working directory: \n\'{cwd}\'\n')
    print(f'Please wait while synchronizing {n_files} files...')

    for i, [fname, fhash] in enumerate(zip(fname_list, fhash_list)):

        # do not download again if already present  
        if fname in os.listdir():        
            print(f'({i+1}) \'{fname}') 

        # download    
        else: 
            url = bucket_url + fname 
            fname = wget.download(url, bar=_bar_progress)
            is_ok = _check_filehash(fname, fhash)
            if is_ok: 
                print(f' ({i+1}) \'{fname}') 
            else: 
                print('Warning! Downloaded file contains errors.') 
                print('Automatically deleting downloaded file for safety reasons!')
                os.remove(fname)
                
    print('Ready!')

    
def _check_filehash(fname, fhash): 
    '''Check file integrity.'''
    
    sha1 = filehash.FileHash('sha1')
    is_ok = sha1.hash_file(fname) == fhash 
    
    return is_ok 



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"download": "10_Download-demo-data.ipynb",
         "read_datasets": "20_Exploring-the-Crono-MA-XRF-HDF5-file-format.ipynb",
         "report": "20_Exploring-the-Crono-MA-XRF-HDF5-file-format.ipynb",
         "peek_inside": "20_Exploring-the-Crono-MA-XRF-HDF5-file-format.ipynb",
         "gridify": "30_Gridify-by-selection.ipynb",
         "convert_crono": "40_Interpolation-gymnastics.ipynb",
         "tree": "40_Interpolation-gymnastics.ipynb",
         "get_array": "40_Interpolation-gymnastics.ipynb",
         "MAXRF_SPECTRA": "40_Interpolation-gymnastics.ipynb",
         "MAXRF_CUBE": "40_Interpolation-gymnastics.ipynb",
         "MAXRF_MAX_SPECTRUM": "40_Interpolation-gymnastics.ipynb",
         "MAXRF_SUM_SPECTRUM": "40_Interpolation-gymnastics.ipynb",
         "MAXRF_ENERGIES": "40_Interpolation-gymnastics.ipynb"}

modules = ["data.py",
           "crono.py",
           "interpolation.py"]

doc_url = "https://openheritagescience.org/cronomaxrf/"

git_url = "https://github.com/fligt/cronomaxrf/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/40_Interpolation-gymnastics.ipynb.

# %% auto 0
__all__ = ['MAXRF_SPECTRA', 'MAXRF_CUBE', 'MAXRF_MAX_SPECTRUM', 'MAXRF_SUM_SPECTRUM', 'MAXRF_ENERGIES', 'convert_crono', 'tree',
           'get_array']

# %% ../notebooks/40_Interpolation-gymnastics.ipynb 28
# python package for reading hdf5 files 
import h5py 

# python package for processing too-big-for-memory data 
import dask 
import dask.array as da 
import dask_ndfilters 
from dask.diagnostics import ProgressBar 

# python package for a new data container file format called z-arrays
import zarr 

# array computions 
import numpy as np 

# regular expressions
import re 

# plotting and printing 
import matplotlib.pyplot as plt 
from IPython.display import HTML 

# %% ../notebooks/40_Interpolation-gymnastics.ipynb 29
# standard datapath locations for zarr zipstore 
MAXRF_SPECTRA = 'maxrf_all_spectra'
MAXRF_CUBE = 'maxrf_cube'
MAXRF_MAX_SPECTRUM = 'maxrf_max_spectrum'
MAXRF_SUM_SPECTRUM = 'maxrf_sum_spectrum'
MAXRF_ENERGIES = 'maxrf_energies'


# FUNCTION TO READ AND CONVERT CRONO HDF5 FILE 

def convert_crono(crono_filename): 
    '''Read spectral data from Crono maxrf hdf5 file and convert to zarr zipstore format.
    
    Compose (smoothed) spectral image data cube from selected spectra. Furthermore, 
    calculate smoothed sum and max spectra. '''
    
    
    with h5py.File(crono_filename, mode='r') as fh: 

        # STEP 1: SCHEDULE COMPUTIONS WITH (LAZY) DASK ARRAYS 

        # read spectral data cube from hdf5 dataset into dask array 
        dataset = fh['/XRF/Spectra']
        arr = da.from_array(dataset) 
        arr = arr.astype(np.float32)

        # schedule spectral dimension gaussian smoothing computation 
        spectra_arr = dask_ndfilters.gaussian_filter(arr, (0, 7))

        n_spectra, n_channels = spectra_arr.shape  

        # READ SCAN POSITIONS AND SHUFFLE SPECTRA INTO CUBE 

        # read and convert position indices to numpy array and remove inner bracket 
        spectra_indices = fh['/XRF/SpectraSelectedIndex'][:,:,0]

        # spectral image height and width  
        h, w = spectra_indices.shape 

        # flatten indices table into list
        spectra_indices = spectra_indices.flatten()

        # shuffle spectrum positions according to indices 
        cube_arr = spectra_arr[spectra_indices].reshape([h, w, n_channels]) 

        # read energy (keV) calibration values of channels  
        energy_vector = fh['/XRF/EnergyVector'][:] 
        # find channel index of our favorite element iron 
        FeKa_i = np.argmin((energy_vector - 6.402)**2)

        # STEP 2: EXECUTE DASK ARRAY COMPUTIONS AND WRITE TO ZIPSTORE 
             
        # create/overwrite and open an empty zarr zipstore file for writing 
        zs_filename = re.sub('\.[^.]*$', '.zipstore', crono_filename) 
        zs = zarr.ZipStore(zs_filename, mode='w') 

        # compute and write maxrf data to datapath in zipstore 
        
        print('Computing (smoothed) spectra...')
        with ProgressBar(): 
            spectra_arr.to_zarr(zs, component=MAXRF_SPECTRA)  
        
        print('Computing (smoothed) spectral data cube from selection...')
        with ProgressBar(): 
            cube_arr.to_zarr(zs, component=MAXRF_CUBE)  
            
        print('Computing (smoothed) max spectrum...')
        with ProgressBar(): 
            (cube_arr.reshape([h*w, n_channels])).max(axis=0).to_zarr(zs, component=MAXRF_MAX_SPECTRUM)

        print('Computing (smoothed) sum spectrum...')
        with ProgressBar():
            ((cube_arr.reshape([h*w, n_channels])).sum(axis=0) / (h*w) ).to_zarr(zs, component=MAXRF_SUM_SPECTRUM)
        
        print('Writing channel energies...')
        with ProgressBar(): 
            da.from_array(energy_vector).to_zarr(zs, component=MAXRF_ENERGIES)
            
        zs.close()
            
        return zs_filename 


# FUNCTIONS FOR USING ZARR ZIPSTORE FILES 
    
def tree(zs_filename, show_arrays=False): 
    '''Prints content tree of *zipstore_file*'''

    with zarr.ZipStore(zs_filename, mode='r') as zs: 
        root = zarr.group(store=zs) 
        tree = root.tree(expand=True).__repr__()
        print(f'\n{zs_filename}:\n\n{tree}')  
        
        if show_arrays:        
            datasets = sorted(root)
            arrays_html = ''

            for ds in datasets: 
                arr = da.from_array(root[ds])
                html = arr._repr_html_()
                arrays_html = f'{arrays_html}- Dataset: <h style="color:brown">{ds}</h>{html}' 
   
            return HTML(arrays_html)


def get_array(zipstore_filename, datapath, compute=True): 
    '''Open zipstore *zipstore_filename* and return dataset from *datapath*. 
    
    For large files that do not fit into memory it is advised to set option compute=False. '''
    
    # open existing zipstore filehandle 
    zs = zarr.ZipStore(zipstore_filename, mode='r') 
    root = zarr.group(store=zs)
    
    # initialize dask array 
    arr = da.from_array(root[datapath])
    
    # return numpy array and close, 
    # otherwise return dask array and do not close  
    if compute: 
        arr = arr.compute()
        zs.close()
        
    return arr 



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/cronomaxrf',
                'doc_host': 'https://fligt.github.io',
                'git_url': 'https://github.com/fligt/cronomaxrf',
                'lib_path': 'cronomaxrf'},
  'syms': { 'cronomaxrf.crono': { 'cronomaxrf.crono._get_attrs': ( 'exploring-the-crono-ma-xrf-hdf5-file-format.html#_get_attrs',
                                                                   'cronomaxrf/crono.py'),
                                  'cronomaxrf.crono._nesting': ( 'exploring-the-crono-ma-xrf-hdf5-file-format.html#_nesting',
                                                                 'cronomaxrf/crono.py'),
                                  'cronomaxrf.crono.peek_inside': ( 'exploring-the-crono-ma-xrf-hdf5-file-format.html#peek_inside',
                                                                    'cronomaxrf/crono.py'),
                                  'cronomaxrf.crono.read_datasets': ( 'exploring-the-crono-ma-xrf-hdf5-file-format.html#read_datasets',
                                                                      'cronomaxrf/crono.py'),
                                  'cronomaxrf.crono.report': ( 'exploring-the-crono-ma-xrf-hdf5-file-format.html#report',
                                                               'cronomaxrf/crono.py')},
            'cronomaxrf.data': { 'cronomaxrf.data._bar_progress': ('download-demo-data.html#_bar_progress', 'cronomaxrf/data.py'),
                                 'cronomaxrf.data._check_filehash': ('download-demo-data.html#_check_filehash', 'cronomaxrf/data.py'),
                                 'cronomaxrf.data.download': ('download-demo-data.html#download', 'cronomaxrf/data.py')},
            'cronomaxrf.gridify': {'cronomaxrf.gridify.gridify': ('gridify-by-selection.html#gridify', 'cronomaxrf/gridify.py')},
            'cronomaxrf.interpolation': { 'cronomaxrf.interpolation.convert_crono': ( 'interpolation-gymnastics.html#convert_crono',
                                                                                      'cronomaxrf/interpolation.py'),
                                          'cronomaxrf.interpolation.get_array': ( 'interpolation-gymnastics.html#get_array',
                                                                                  'cronomaxrf/interpolation.py'),
                                          'cronomaxrf.interpolation.tree': ( 'interpolation-gymnastics.html#tree',
                                                                             'cronomaxrf/interpolation.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/20_Exploring-the-Crono-MA-XRF-HDF5-file-format.ipynb.

# %% auto 0
__all__ = ['read_datasets', 'report', 'peek_inside']

# %% ../notebooks/20_Exploring-the-Crono-MA-XRF-HDF5-file-format.ipynb 17
import h5py 
import numpy as np 

def read_datasets(crono_filename): 
    '''Read all datasets in `crono_filename`.'''
    
    fh = h5py.File(crono_filename, mode='r')

    groups_and_datasets = []
    fh.visit(groups_and_datasets.append) 

    datasets = [fh[d] for d in groups_and_datasets if (type(fh[d]) == h5py.Dataset)] 
    
    return datasets


def report(crono_filename): 
    '''Print info about structure and content of datasets in `crono_filename`.'''
    
    datasets = read_datasets(crono_filename) 
    
    # print list of datasets 
    print(f'CONTENTS OF CRONO HDF5 FILE: \'{crono_filename}\'')
    print(f'({len(datasets)} DATASETS)\n')
    
    for i, d in enumerate(datasets):
        print(f'[{i}] {d.name}')
        
    print()
    
    ruler = '-'*80
    
    for i, d in enumerate(datasets): 
        
        
        print(ruler)
        print(f'[{i}] ', end='')
        peek_inside(d)
        



def peek_inside(dataset, _print=True):
    '''Summarize structure and content of a `dataset`. '''
        
    shape_str, dtype, sub_dtype = _nesting(dataset)


    # simplify represention for certain complex dataset structures 
    if dtype == np.dtype('S1'): 
        value_str = ''.join(dataset[...].astype(str))

    elif sub_dtype == np.dtype('V1'): 
        value_str = bytes(dataset[...][0])

    elif sub_dtype == np.dtype('S1'): 
        str_list = [''.join(s.astype(str)) for s in dataset[...]]
        value_str = '\n'.join(str_list)

    # otherwise use default repr string
    else: 
        value_str = repr(dataset[...])

    # compose summary 
    name = dataset.name
    attributes = _get_attrs(dataset) 
    
    summary = f'{name}:\n\n{attributes}\n{shape_str}\n\n+VALUES: \n\'{value_str}\'\n\n' 

    if _print:  
        print(summary)
        return None
    else: 
        return summary 
    
        
def _nesting(dataset): 
    '''Report shape or nested shapes of numpy arrays in `dataset`.'''
    
    v = dataset[...]
    shape = v.shape 
    dtype = v.dtype 
    
    shape_str = f'+SHAPE: {v.shape} DTYPE: \'{v.dtype}\'' 
    
    sub_dtype = None 
    
    if dtype == np.dtype('O'):
        
        subshape_list = [a.shape for a in v] 
        sub_dtype = v[-1].dtype 
        n_sub = len(subshape_list)
        
        if n_sub > 4: 
            
            subshape_list = subshape_list[0:4]
            subshape_list.append(f'....')
        
        subshape_str = f'{subshape_list}' 
        shape_str = shape_str + ' SUBSHAPES: ' + subshape_str + f' SUB_DTYPE: \'{sub_dtype}\''
    
    return shape_str, dtype, sub_dtype 


def _get_attrs(dataset): 
    '''Report information from `dataset` attributes.'''
    
    attr_items = dataset.attrs.items()
    attr_keys = []
    attr_values = []

    for k, v in attr_items:
        attr_keys.append(k) 

        # get rid of outer array 
        if v[0].dtype == np.dtype('S1'):       
            v = ''.join(v[0].astype(str))  
        if type(v) == np.ndarray: 
            v = v[0]
        
        if k == 'MapSetup': 
            v = bytes(v)
        
        if k == 'TubeTemperature': 
            v = f'{v:0.2f}'
            
        attr_values.append(v) 
            

    attr_dict = dict(zip(attr_keys, attr_values))
    
    if len(attr_dict) > 0: 
    
        repr_string = '+ATTRIBUTES: \n'

        for k, v in attr_dict.items():

            repr_string = repr_string + f'        - {k}: {v}\n'
            
    else: 
        repr_string = '+ATTRIBUTES: (none)\n'
        
    #print(repr_string) 
        
   
    return repr_string



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/30_Gridify-by-selection.ipynb.

# %% auto 0
__all__ = ['gridify']

# %% ../notebooks/30_Gridify-by-selection.ipynb 25
# python package for reading hdf5 files 
import h5py 

# python package for processing too-big-for-memory data 
import dask 
import dask.array as da 
#import dask_ndfilters 
from dask.diagnostics import ProgressBar 

# standard imports  
import re 
import os

def gridify(crono_filename): 
    '''Export Crono maxrf file spectral data into regular spectral image hdf5 file. '''
    
    # Datamunger should be able to read these 
    MAXRF_IMAGE = '/spectra'
    MAXRF_ENERGIES = '/wavelength'

    with h5py.File(crono_filename, mode='r') as fh: 

        # read spectra and energies from hdf5 dataset into (lazy) dask arrays  
        spectra = fh['/XRF/Spectra']
        dask_spectra = da.from_array(spectra) 
        energies = fh['/XRF/EnergyVector']
        dask_energies = da.from_array(energies)

        # load selected indices into memory as numpy array 
        selected = fh['/XRF/SpectraSelectedIndex'][:,:,0]  
        dask_gridified = dask_spectra.vindex[selected] # dask (lazy) fancy indexing 

        # create filename for saving to 
        ptrn = '\.[^\.]*$'  
        repl = '_GRIDIFIED.HDF5'
        gridified_filename = re.sub(ptrn, repl, crono_filename)

        # write gridified spectral image to hdf5 file 
        print(f'Converting \'{crono_filename}\':\n')
        
        print(f'(1/2) Writing dataset \'{MAXRF_IMAGE}\' with shape {dask_gridified.shape} to hdf5 file...') 
        with ProgressBar(): 
            dask_gridified.to_hdf5(gridified_filename, MAXRF_IMAGE)

        # also write channel energies to hdf5 file 
        print(f'(2/2) Writing dataset \'{MAXRF_ENERGIES}\' with shape {dask_energies.shape} to hdf5 file...')
        with ProgressBar(): 
            dask_energies.to_hdf5(gridified_filename, MAXRF_ENERGIES)
            
        filesize_MB = os.path.getsize(gridified_filename) // 1e6
        print(f'File size: {filesize_MB} MB')
            
        return gridified_filename

